
usePackage <- function(p) {
  if (!is.element(p, installed.packages()[,1]))
    install.packages(p, dep = TRUE)
  require(p, character.only = TRUE)
}

usePackage("glmnet")
usePackage("pROC")
usePackage("caret")
usePackage("ROCR")
usePackage("MLmetrics")
usePackage("Matrix")


#balancing matrices when using cBind, NA rows added to matrix with minor rows
bal_matrix_rows<-function(matrix_less, matrix_more) {
  
  add<-length(matrix_more[,1])-length(matrix_less[,1])
  output_matrix<-rBind(matrix_less, Matrix(NA,ncol=length(colnames(matrix_less)),nrow=add,sparse=T))
  
  return(output_matrix)
  
}  

## Evaluate LASSO (via GLMNET) with PMAX
evalGLMNET_PMAX <- function(X, Y, Xt, Yt, alpha, dataset_name, fold_num, rnd_seed, pen, pmax, weights) {
  
  #check if any columns w/ positive cases and drop them
  removed_zero<-intersect(colnames(X)[which(colSums(X)==0)],colnames(Xt)[which(colSums(Xt)==0)])

  ptm <- proc.time()
  # build model
  fit <- NULL
  #set.seed(rnd_seed)
  if(is.null(pmax)) {
    try(fit <- cv.glmnet(X[,setdiff(colnames(X),removed_zero)], Y, family="binomial", type.measure = "auc", weights = weights,  nfolds=5, alpha=alpha, penalty.factor=pen))
  } else {
    try(fit <- cv.glmnet(X[,setdiff(colnames(Xt),removed_zero)], Y, family="binomial", type.measure = "auc", weights = weights, nfolds=5, alpha=alpha, penalty.factor=pen, pmax=pmax))
    
  }
  if(is.null(fit)) return(NULL)
  #time <- proc.time()-ptm
  #print(time)
  
  
  # Count selected features
  numSelectedMin <- fit$nzero[which(fit$lambda == fit$lambda.min)]

  #needed for threshold
  predMin <- predict(fit, type="response", X[,setdiff(colnames(X),removed_zero)], s="lambda.min")
  
  #predict on test
  predMin_test <- predict(fit, type="response", Xt[,setdiff(colnames(Xt),removed_zero)], s="lambda.min")

  brier_test <- mean((predMin_test-Yt)^2)
  
  # Calculate threshold on training data, other metrics (except AUC,AUPRC)
  # are calculated on this optimal fixed threshold
  result.roc <- roc(Y, as.numeric(predMin))
  thres<- coords(result.roc, "best", best.method="closest.topleft", ret=c("threshold"))
  
  #more then 1 threshold, exmaple -Inf, + Inf, only -Inf kept
  if(length(thres)!=1) thres<-thres[1]

  #predicting classes on test set
  predMin_test_class<-(predMin_test > thres )*1
  
 
  #output for testing
  #confusionMatrix(Yt, (predMin_test > thres )*1, positive="1")
  
  #AUC and AUPRC on test set  
  auc_test<-auc(Yt, as.vector(predMin_test))
  auprc_test <- PRAUC(y_pred = as.numeric(predMin_test), y_true = Yt)
  
  percPos_test <- sum(predMin_test_class)/length(predMin_test_class)
  
  #output for variables
  c<-coef(fit,s='lambda.min',exact=TRUE)

  indsPos<-which(c>0)
  indsNeg<-which(c<0)
  variables<-row.names(c)[indsPos]
  
  if(length(variables[!variables %in% '(Intercept)']) >0) {
  variables<-paste0("+",variables[!variables %in% '(Intercept)'])
  } else {variables<-NULL}
  
  temp<-row.names(c)[indsNeg]
  if(length(temp) > 0) {
    variables<-c(variables, paste0("-",temp[!temp %in% '(Intercept)']))
  }
  variables<-paste(variables,sep="",collapse=",")
  variables<-gsub(" ",",",variables)

  output<-data.frame(Name=dataset_name,Model="Normal_one_stage",Fold=fold_num, Seed=rnd_seed,
                     PMAX=ifelse(is.null(pmax), "no_pmax",pmax),
                     AUC=as.numeric(auc_test),
                     AUPRC=as.numeric(auprc_test),
                     Threshold= thres,Brier=brier_test,
                     Accuracy=Accuracy(predMin_test_class, Yt),
                     Sensitivity=sensitivity(factor(predMin_test_class),factor(Yt),positive="1"),
                     Specificity=specificity(factor(predMin_test_class),factor(Yt),negative="0"),
                     PPV=posPredValue(factor(predMin_test_class),factor(Yt),positive="1"),
                     NPV=negPredValue(factor(predMin_test_class),factor(Yt),negative="0"),
                     NumSelected=numSelectedMin,
                     Zero_pred=length(removed_zero),YLength=length(Y), YtLength=length(Yt),
                     YSum=sum(Y), YtSum=sum(Yt),
                     PercPos=percPos_test,
                     Pred_Vars=variables, stringsAsFactors = F)

  #output for calibration plots (predicted probability on Yt, Yt)
  output_calib<-Matrix(cbind(predMin_test,Yt),sparse=T)

  #print all but perdictor variables
  print(output[,1:(length(names(output))-1)])
  
  time_full <- proc.time()-ptm
  print(time_full)
  return(list(output,output_calib ))
}
